# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
SGP4 Propagator to propagate Earth bound satellites.

"""
import numpy as np
from astropy import units as u
from astropy.coordinates import (
    GCRS,
    TEME,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)

from satmad.coordinates.trajectory import Trajectory
from satmad.core.celestial_bodies import EARTH
from satmad.propagation.base_propagator import AbstractPropagator


class SGP4Propagator(AbstractPropagator):
    """
    SGP4 Analytical propagator using `TEME` mean orbital elements.

    While `central_body` of the propagator is `EARTH`, its constants are not used.

    Parameters
    ----------
    name : str
        Name of the propagator
    stepsize : Quantity or TimeDelta
        output stepsize for the propagator
    """

    def __init__(self, stepsize=60 * u.s, name="SGP4"):
        super().__init__(stepsize, name, central_body=EARTH)

    @staticmethod
    def propagate(tle, time):
        """
        Computes the coordinates at the target `time` using the source `TLE` mean
        orbital elements.

        Parameters
        ----------
        tle : TLE
            Mean orbital elements (Two-Line-Element)
        time : Time
            Target time for propagation

        Returns
        -------
        coords : SkyCoord
            Coordinates at target time (in `GCRS`)

        Raises
        ------
        SGP4GeneralError
            Errors in SGP4 propagation
        SGP4SatDecayedError
            Satellite coordinates at required coordinates are below decay altitude.
        """
        # compute coords at time
        e, r, v = tle.satrec.sgp4(time.utc.jd1, time.utc.jd2)

        # check error code
        SGP4Propagator.__handle_err_code(e)

        v_teme = CartesianDifferential(np.asarray(v), unit=u.km / u.s)
        r_teme = CartesianRepresentation(np.asarray(r), unit=u.km)
        rv_gcrs = TEME(r_teme.with_differentials(v_teme), obstime=time).transform_to(
            GCRS(obstime=time)
        )

        return SkyCoord(
            rv_gcrs, representation_type="cartesian", differential_type="cartesian"
        )

    def gen_trajectory(self, tle, interval):
        """
        Generates the trajectory (time and coordinate information) for the given
        interval with the internal stepsize.
        Parameters
        ----------
        tle : TLE
            Two-Line-Element initial orbit information (TEME mean orbital elements)
        interval : TimeInterval
            Time interval for which the ephemeris will be generated
        Returns
        -------
        Trajectory
            The output trajectory (in `GCRS`)
        """

        # ****** Generate the pos, vel vectors for each time instant ******

        # generate the output timelist
        time_list = self._generate_time_list(interval)

        # Run the propagation and init pos and vel vectors in TEME
        e, r_list, v_list = tle.satrec.sgp4_array(time_list.jd1, time_list.jd2)

        # Load the time, pos, vel info into astropy objects (shallow copied)
        rv_list_teme = CartesianRepresentation(
            r_list, unit=u.km, xyz_axis=1
        ).with_differentials(CartesianDifferential(v_list, unit=u.km / u.s, xyz_axis=1))

        rv_list_gcrs = TEME(
            rv_list_teme,
            obstime=time_list,
            representation_type="cartesian",
            differential_type="cartesian",
        ).transform_to(GCRS(obstime=time_list))

        # trajectory in astropy
        traj_astropy = SkyCoord(
            rv_list_gcrs,
            obstime=time_list,
            frame="gcrs",
            representation_type="cartesian",
            differential_type="cartesian",
        )

        # Init trajectory in Trajectory object
        trajectory = Trajectory(traj_astropy)

        return trajectory

    @staticmethod
    def __handle_err_code(e):
        """
        Handle the error code out of the propagation.

        Error codes are:
            0. no error
            1. mean elements, ecc >= 1.0 or ecc < -0.001 or a < 0.95 earth radius
            2. mean motion less than 0.0
            3. perturbed eccentricity out of limits, ecc < 0.0  or  ecc > 1.0
            4. semi-latus rectum < 0.0
            5. epoch elements are sub-orbital
            6. satellite has decayed

        Parameters
        ----------
        e : int
            error code

        Raises
        ------
        SGP4GeneralError
            Errors in SGP4 propagation - satellite orbit evolved to an invalid state.
        SGP4SatDecayedError
            Satellite coordinates at required coordinates are below decay altitude.
        """
        if e == 0:
            pass
        if e == 1:
            raise SGP4GeneralError(
                "SGP4 Propagation Error: "
                "ecc >= 1.0 or ecc < -0.001 or a < 0.95 earth radius"
            )
        if e == 2:
            raise SGP4GeneralError("SGP4 Propagation Error: mean motion less than 0.0")
        if e == 3:
            raise SGP4GeneralError(
                "SGP4 Propagation Error: "
                "perturbed eccentricity out of limits, ecc < 0.0  or  ecc > 1.0"
            )
        if e == 4:
            raise SGP4GeneralError("SGP4 Propagation Error: semi-latus rectum < 0.0")
        if e == 5:
            raise SGP4GeneralError(
                "SGP4 Propagation Error: epoch elements are sub-orbital"
            )
        if e == 6:
            raise SGP4SatDecayedError(
                "SGP4 Propagation Error:"
                "Satellite has decayed (Sat number: {tle.sat_number})"
            )


class SGP4GeneralError(Exception):
    """General SGP4 Propagation errors, usually orbital elements out of range."""

    pass


class SGP4SatDecayedError(Exception):
    """Satellite coordinates at required coordinates are below decay altitude."""

    pass



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2021 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
TLE storage classes.

"""
from abc import ABC
from copy import copy
from enum import Enum
from typing import List

from satmad.propagation.tle import TLE


class TleFilterParams(Enum):
    """TLE Filtering Parameters."""

    PERIOD = "period"
    SM_AXIS = "sm_axis"
    NODE_ROTATION_RATE = "node_rotation_rate"
    ARGP_ROTATION_RATE = "argp_rotation_rate"
    EPOCH = "epoch"
    BSTAR = "bstar"
    N_DOT = "n_dot"
    N_DOTDOT = "n_dotdot"
    INCLINATION = "inclination"
    RAAN = "raan"
    ECCENTRICITY = "eccentricity"
    ARG_PERIGEE = "arg_perigee"
    MEAN_ANOMALY = "mean_anomaly"
    MEAN_MOTION = "mean_motion"
    NAME = "name"
    SAT_NUMBER = "sat_number"
    INTL_DESIGNATOR = "intl_designator"
    CLASSIFICATION = "classification"
    REV_NR = "rev_nr"
    TLE = "tle"


class _TleList(ABC):
    """Abstract Base Class for TLE lists."""

    tle_list: List[TLE] = []

    def filter_by_value(self, param, value):
        """
        Filters the TLE list for equivalence to a given value.

        For example `param` can be equal to `TleFilterParams.SAT_NUMBER` and `value`
        can be equal to `46945`, which filters all TLEs with a sat number 46945.

        Note that this filter is not appropriate for float values such as eccentricity
        where equivalence is very brittle. For these applications, use
        `filter_by_range()` or `filter_by_func()` instead.

        This method returns a `TleStorage` object even if the filtering result is empty.
        In this case, `tle_list` parameter of the `TleStorage` object will be an empty
        list.

        Also note that, the returned TLE objects in the list are just shallow copies
        of the objects in the master TLE list. Any change to them will change the
        relevant item in the backing TLE list.

        Parameters
        ----------
        param : TleFilterParams
            Filter parameter (such as name or satellite number)
        value
            Value associated with the parameter

        Returns
        -------
        TleStorage
            A `TleStorage` object that contains the filtered list of TLE data

        Raises
        ------
        ValueError
            TleFilterParams.TLE is given as an input
        """
        if param is TleFilterParams.TLE:
            raise ValueError(
                "TleFilterParams.TLE is not defined for filtering by value."
            )

        filtered_list = [
            tle for tle in self.tle_list if getattr(tle, param.value) == value
        ]

        # create new object with the filtered list
        return self._selfcopy(filtered_list)

    def filter_by_func(self, param, filter_func):
        """
        Filters the TLE list for compliance to a given filter function.

        For example `param` can be equal to `TleFilterParams.SM_AXIS` and
        `filter_func` can be a filtering function that tests this parameter and
        returns `True` or `False` accordingly. The following function filters for
        semi-major axis values above 7000 km. Note that units should be defined and
        compatible with the value to be compared against. For semimajor axis, distance
        units such as meters are acceptable but using no dimensions or using
        wrong units (such as degrees) will throw an error.

        >>> from astropy import units as u
        >>> def sma_filter(a):
        >>>     return True if a > 7000 * u.km else False

        For exact equivalences (such as satellite names or ID numbers),
        using `filter_by_value` method will be easier and more appropriate.

        This method returns a `TleStorage` object even if the filtering result is empty.
        In this case, `tle_list` parameter of the `TleStorage` object will be an empty
        list.

        Also note that, the returned TLE objects in the list are just shallow copies
        of the objects in the master TLE list. Any change to them will change the
        relevant item in the backing TLE list.

        Parameters
        ----------
        param : TleFilterParams
            Filter parameter (such as name or satellite number)
        filter_func
            Function to test the parameter against

        Returns
        -------
        TleStorage
            A `TleStorage` object that contains the filtered list of TLE data
        """
        filtered_list = [
            tle for tle in self.tle_list if filter_func(getattr(tle, param.value, tle))
        ]

        # create new object with the filtered list
        return self._selfcopy(filtered_list)

    def filter_by_range(self, param, min_value=None, max_value=None):
        """
        Filters the TLE list for compliance to a given min/max values.

        The test is simply:

        `max_value > param > min_value`

        For example `param` can be equal to `TleFilterParams.SM_AXIS`, then
        this parameter is tested against the minimum and maximum semimajor axis values
        supplied. If `None` is supplied for `min_value` or `max_value`, then there
        is no range or range check defined for this parameter. For example,
        if `min_value` is `None`, the parameter check reduces to `max_value > param`.

        Note that units should be defined and
        compatible with the value to be compared against. For semimajor axis, distance
        units such as meters are acceptable but using no dimensions or using
        wrong units (such as degrees) will throw an error.

        For exact equivalences (such as satellite names or ID numbers),
        using `filter_by_value` method will be easier and more appropriate.

        This method returns a `TleStorage` object even if the filtering result is empty.
        In this case, `tle_list` parameter of the `TleStorage` object will be an empty
        list.

        Also note that, the returned TLE objects in the list are just shallow copies
        of the objects in the master TLE list. Any change to them will change the
        relevant item in the backing TLE list.

        Parameters
        ----------
        param : TleFilterParams
            Filter parameter (such as name or satellite number)
        min_value
            Minimum value to test the parameter against
        max_value
            Maximum value to test the parameter against

        Returns
        -------
        TleStorage
            A `TleStorage` object that contains the filtered list of TLE data

        Raises
        ------
        ValueError
            TleFilterParams.TLE is given as an input
        """
        if param is TleFilterParams.TLE:
            raise ValueError(
                "TleFilterParams.TLE is not defined for filtering by value."
            )

        # `min_value` and `max_value` are quantities and should be checked explicitly
        # for `None`, otherwise can be interpreted as `True` or `False`.
        if min_value is not None and max_value is not None:
            filtered_list = [
                tle
                for tle in self.tle_list
                if max_value > getattr(tle, param.value) > min_value
            ]
        elif min_value is not None:
            filtered_list = [
                tle
                for tle in self.tle_list
                if getattr(tle, param.value, None) > min_value
            ]
        elif max_value is not None:
            filtered_list = [
                tle
                for tle in self.tle_list
                if max_value > getattr(tle, param.value, None)
            ]
        else:
            filtered_list = []

        # create new object with the filtered list
        return self._selfcopy(filtered_list)

    def _selfcopy(self, new_list):
        """Creates a new (shallow copied) object of the same type with the new list."""
        output = copy(self)
        output.tle_list = new_list

        return output


class TleTimeSeries(_TleList):
    """TLE storage class that keeps a list of TLE data from a single satellite,
    at multiple times and with time order.

    The entry point is ideally the `TleStorage` class, where a TLE file is usually read,
    and a single satellite is filtered. Once this class is initialised, various sublists
    (e.g. specific time range) can be derived.

    Parameters
    ----------
    tle_list : list[TLE]
        initial list of TLE objects (shallow copied into object)
    """

    def __init__(self, tle_list, sat_number):

        # init a TLE Storage and filter for the sat number
        self.tle_list = (
            TleStorage(tle_list)
            .filter_by_value(TleFilterParams.SAT_NUMBER, sat_number)
            .tle_list
        )

        # order the internal TLE list with respect to epoch
        self.tle_list.sort(key=lambda tle: tle.epoch)


class TleStorage(_TleList):
    """TLE storage class that keeps a list of TLE data from multiple satellites,
    at multiple times and without any ordering.

    This class is the entry point for reading a TLE file, from which various sublists
    (e.g. single satellite, all LEO sats etc.) can be derived.

    Parameters
    ----------
    tle_list : list[TLE]
        initial list of TLE objects (shallow copied into object)
    """

    def __init__(self, tle_list):
        self.tle_list = tle_list

    @classmethod
    def from_path(cls, tle_file_path):
        """
        Read a set of TLE data from file. Tries to extract satellite names from the
        list, if no name is found, an empty string (not `None`) is assigned.

        Parameters
        ----------
        tle_file_path : Path
            Path of the text file to be read

        Returns
        -------
        TleStorage
            A `TleStorage` object that contains the list of TLE data
        """
        with open(tle_file_path, "r") as f:
            tle_source_str = f.read()

        return cls.from_string(tle_source_str)

    @classmethod
    def from_string(cls, tle_string):
        """
        Read a set of TLE data from a text. Tries to extract satellite names from the
        list, if no name is found, an empty string (not `None`) is assigned.

        Parameters
        ----------
        tle_string : str
            String containing successive TLE data

        Returns
        -------
        TleStorage
            A `TleStorage` object that contains the list of TLE data
        """

        tle_source_str = tle_string.split("\n")

        # create object without calling `__init__`
        tle_storage = cls.__new__(cls)

        # parse the string and load parsed items into list
        tle_storage.tle_list = _parse_tle_list(tle_source_str)

        return tle_storage

    def to_tle_timeseries(self, sat_number):
        """
        Filters the TLE list for a single satellite to initialise a `TleTimeSeries`.

        Parameters
        ----------
        sat_number
            Satellite Catalog Number

        Returns
        -------
        TleTimeSeries
            A `TleTimeSeries` object that contains the list of TLE data of a single
            satellite

        """
        return TleTimeSeries(self.tle_list, sat_number)


def _parse_tle_list(tle_source_str_list):
    """
    Parses the TLE list.

    Parameters
    ----------
    tle_source_str_list : list[str]
        TLE data as a list of strings.

    Returns
    -------
    tle_list : list[TLE]
        List of TLE data
    """

    tle_list = []

    name = line1 = line2 = None
    for i, line in enumerate(tle_source_str_list):

        # strip spaces and EOF around the line
        line = line.strip()
        # skip empty lines
        if not line.strip():
            continue

        if __is_tle_line(line, 1):
            line1 = line
            if __is_tle_line(tle_source_str_list[i + 1], 2):
                line2 = tle_source_str_list[i + 1]
                if i > 0 and (
                    not __is_tle_line(tle_source_str_list[i - 1], 1)
                    and not __is_tle_line(tle_source_str_list[i - 1], 2)
                ):
                    name = tle_source_str_list[i - 1].strip("\n ")
                    if name.startswith("0 "):
                        name = name[2:]

        if line1 and line2:
            tle = TLE.from_tle(line1, line2, name if name else "")
            tle_list.append(tle)
            # reset temp fields
            name = line1 = line2 = None

    return tle_list


def __is_tle_line(line, line_nr):
    """Checks whether line is of type Line 1 or Line 2 standard TLE line."""
    if line.strip().startswith(f"{line_nr} "):
        return True
    else:
        return False



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Test `TimeInterval` class and associated methods and functionalities.

"""
import numpy as np
import pytest
from astropy import units as u
from astropy.time import Time, TimeDelta

from satmad.utils.timeinterval import _EPS_TIME, TimeInterval, TimeIntervalList

before = TimeInterval(
    Time("2020-04-09T00:00:00", scale="utc"),
    Time("2020-04-11T00:00:00", scale="utc"),
    end_inclusive=False,
)
within = TimeInterval(
    Time("2020-04-11T00:05:00", scale="utc"), Time("2020-04-11T00:08:00", scale="utc")
)
intersect = TimeInterval(
    Time("2020-04-10T00:00:00", scale="utc"), Time("2020-04-11T00:08:00", scale="utc")
)
exact = TimeInterval(
    Time("2020-04-11T00:00:00", scale="utc"), Time("2020-04-11T00:10:00", scale="utc")
)
after = TimeInterval(
    Time("2020-04-11T00:10:00", scale="utc"), Time("2020-04-12T00:00:00", scale="utc")
)


@pytest.fixture
def init_times():
    """Generates the initial times"""
    return Time("2020-04-10T00:00:00", scale="utc") + np.arange(1, 4) * TimeDelta(
        1, format="jd"
    )


@pytest.fixture
def durations():
    """Generates the initial durations for the initial times"""
    return np.arange(1, 4) * TimeDelta(600, format="sec")


def test_interval_init_with_copy(init_times, durations):
    """Test initialisation with copy."""
    interval_1 = TimeInterval(init_times, durations)
    interval_2 = TimeInterval(interval_1, None)
    interval_3 = TimeInterval(interval_1, None, replicate=True)

    assert interval_1.is_equal(interval_2)
    assert interval_1.is_equal(interval_3)


def test_interval_init_with_durations(init_times, durations):
    """Test initialisation with durations."""
    interval = TimeInterval(init_times, durations)

    truth_txt = "[ 2020-04-11T00:00:00.000  2020-04-11T00:10:00.000 ]"
    assert truth_txt == str(interval)


def test_interval_init_with_end_times(init_times, durations):
    """Test initialisation with explicit end times."""
    end_times = init_times + durations

    interval = TimeInterval(init_times, end_times, replicate=True, end_inclusive=False)

    truth_txt = "[ 2020-04-11T00:00:00.000  2020-04-11T00:10:00.000 )"
    assert truth_txt == str(interval)


def test_interval_init_switched_err(init_times, durations):
    """Test `init` with switched init and end times - should raise `ValueError`."""
    with pytest.raises(ValueError):
        end_times = init_times + durations
        TimeInterval(end_times, init_times)


def test_interval_init_zero_dur_err(init_times):
    """Test `init` with equal init and end times - should raise `ValueError`."""
    with pytest.raises(ValueError):
        TimeInterval(init_times, init_times + 0.1 * _EPS_TIME)


def test_interval_list_init_with_durations(init_times, durations):
    """Test initialisation with durations."""

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], durations[i]))

    intervals = TimeIntervalList(init_intervals)

    truth_txt = (
        "[ 2020-04-11T00:00:00.000  2020-04-11T00:10:00.000 ]\n"
        "[ 2020-04-12T00:00:00.000  2020-04-12T00:20:00.000 ]\n"
        "[ 2020-04-13T00:00:00.000  2020-04-13T00:30:00.000 ]\n"
    )

    assert truth_txt == str(intervals)


def test_interval_list_init_with_end_times(init_times, durations):
    """Test initialisation with explicit end times."""

    end_times = init_times + durations

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], end_times[i]))

    intervals = TimeIntervalList(
        init_intervals, start_valid=init_times[0], end_valid=end_times[-1]
    )

    truth_txt = (
        "[ 2020-04-11T00:00:00.000  2020-04-11T00:10:00.000 ]\n"
        "[ 2020-04-12T00:00:00.000  2020-04-12T00:20:00.000 ]\n"
        "[ 2020-04-13T00:00:00.000  2020-04-13T00:30:00.000 ]\n"
    )

    truth_validity = "[ 2020-04-11T00:00:00.000  2020-04-13T00:30:00.000 ]"

    assert truth_txt == str(intervals)
    assert truth_validity == str(intervals.valid_interval)


def test_interval_list_intersection(init_times, durations):
    """Test interval intersection."""

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], durations[i]))

    intervals = TimeIntervalList(init_intervals)

    interval_full_intersect = TimeInterval(Time("2020-04-13T00:00:00.000"), 5 * u.min)
    interval_part_intersect = TimeInterval(Time("2020-04-10T23:50:00.000"), 60 * u.min)
    interval_no_intersect = TimeInterval(Time("2020-04-11T12:00:00.000"), 5 * u.min)

    # *** Test TimeIntervalList vs. TimeInterval intersection ***
    assert intervals.is_intersecting(interval_full_intersect) is True
    assert "[ 2020-04-13T00:00:00.000  2020-04-13T00:05:00.000 ]" == str(
        intervals.intersect(interval_full_intersect)
    )

    assert intervals.is_intersecting(interval_part_intersect) is True
    assert "[ 2020-04-11T00:00:00.000  2020-04-11T00:10:00.000 ]" == str(
        intervals.intersect(interval_part_intersect)
    )

    assert intervals.is_intersecting(interval_no_intersect) is False
    assert intervals.intersect(interval_no_intersect) is None

    # *** Test TimeIntervalList vs. TimeIntervalList intersection ***

    test_intervals = TimeIntervalList(
        [interval_full_intersect, interval_part_intersect, interval_no_intersect]
    )

    truth_txt = (
        "[ 2020-04-11T00:00:00.000  2020-04-11T00:10:00.000 ]\n"
        "[ 2020-04-13T00:00:00.000  2020-04-13T00:05:00.000 ]\n"
    )

    assert truth_txt == str(intervals.intersect_list(test_intervals))


def test_interval_list_union(init_times, durations):
    """Test interval union."""

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], durations[i]))

    intervals = TimeIntervalList(
        init_intervals,
        start_valid=Time("2020-04-11T00:00:00.000"),
        end_valid=Time("2020-04-14T00:00:00.000"),
    )

    interval_part_intersect = TimeInterval(Time("2020-04-10T23:50:00.000"), 60 * u.min)
    interval_no_intersect = TimeInterval(Time("2020-04-11T12:00:00.000"), 5 * u.min)
    interval_full_intersect = TimeInterval(Time("2020-04-13T00:00:00.000"), 5 * u.min)
    interval_outside = TimeInterval(Time("2020-04-13T23:50:00.000"), 15 * u.min)

    # *** Test TimeIntervalList vs. TimeInterval intersection ***
    assert "[ 2020-04-13T00:00:00.000  2020-04-13T00:30:00.000 ]" == str(
        intervals.union(interval_full_intersect)
    )

    assert "[ 2020-04-10T23:50:00.000  2020-04-11T00:50:00.000 ]" == str(
        intervals.union(interval_part_intersect)
    )

    assert intervals.union(interval_no_intersect) is None

    # *** Test TimeIntervalList vs. TimeIntervalList intersection ***

    test_intervals = TimeIntervalList(
        [
            interval_full_intersect,
            interval_part_intersect,
            interval_no_intersect,
            interval_outside,
        ]
    )

    # print(intervals.union_list(test_intervals))
    truth_txt = (
        "[ 2020-04-11T00:00:00.000  2020-04-11T00:50:00.000 ]\n"
        "[ 2020-04-11T12:00:00.000  2020-04-11T12:05:00.000 ]\n"
        "[ 2020-04-12T00:00:00.000  2020-04-12T00:20:00.000 ]\n"
        "[ 2020-04-13T00:00:00.000  2020-04-13T00:30:00.000 ]\n"
        "[ 2020-04-13T23:50:00.000  2020-04-14T00:00:00.000 ]\n"
    )

    assert truth_txt == str(intervals.union_list(test_intervals))


def test_is_intersecting(init_times, durations):
    """Test `is_intersecting` method."""

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], durations[i]))

    interval = TimeIntervalList(init_intervals).get_interval(0)

    assert interval.is_intersecting(before) is False
    assert interval.is_intersecting(within) is True
    assert interval.is_intersecting(intersect) is True
    assert interval.is_intersecting(exact) is True
    assert interval.is_intersecting(after) is False


def test_contains(init_times, durations):
    """Test `contains` method."""

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], durations[i]))

    interval = TimeIntervalList(init_intervals).get_interval(0)

    assert interval.contains(before) is False
    assert interval.contains(within) is True
    assert interval.contains(exact) is True
    assert interval.contains(intersect) is False
    assert interval.contains(after) is False


def test_is_in_interval(init_times, durations):
    """Test `is_in_interval` method."""

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], durations[i]))

    interval = TimeIntervalList(init_intervals).get_interval(0)

    t_before = Time("2020-04-09T00:00:00", scale="utc")
    t_eqinit = Time("2020-04-11T00:00:00", scale="utc")
    t_within = Time("2020-04-11T00:05:00", scale="utc")
    t_eqend = Time("2020-04-11T00:10:00", scale="utc")
    t_after = Time("2020-04-11T12:00:00", scale="utc")

    assert interval.is_in_interval(t_before) is False
    assert interval.is_in_interval(t_eqinit) is True
    assert interval.is_in_interval(t_within) is True
    assert interval.is_in_interval(t_eqend) is True
    assert interval.is_in_interval(t_after) is False


def test_intersect(init_times, durations):
    """Test `intersect` method."""

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], durations[i]))

    interval = TimeIntervalList(init_intervals).get_interval(0)

    assert interval.intersect(before) is None
    assert interval.intersect(within).is_equal(within)
    assert interval.intersect(intersect).is_equal(
        TimeInterval(interval.start, intersect.end)
    )
    assert exact.is_equal(interval.intersect(exact))
    assert exact.is_equal(intersect) is False
    assert interval.intersect(after) is None


def test_union(init_times, durations):
    """Test `union` method."""

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], durations[i]))

    interval = TimeIntervalList(init_intervals).get_interval(0)

    assert interval.union(before) is None
    assert interval.union(within).is_equal(interval)
    assert interval.union(intersect).is_equal(
        TimeInterval(intersect.start, interval.end)
    )
    assert interval.union(exact).is_equal(interval)
    assert interval.union(after) is None


def test_expand():
    """Test `expand` method."""
    # Test expansion
    expanded = within.expand(
        start_delta=TimeDelta(5 * 60, format="sec"),
        end_delta=TimeDelta(2 * 60, format="sec"),
    )
    assert expanded.is_equal(exact)

    # Test shrinkage
    shrunk = exact.expand(
        start_delta=TimeDelta(-5 * 60, format="sec"),
        end_delta=TimeDelta(-2 * 60, format="sec"),
    )
    assert shrunk.is_equal(within)


def test_expand_shrink_zero():
    """Test `expand` method with shrink to zero - should raise `ValueError`."""
    with pytest.raises(ValueError):
        # Test shrink to zero - this raises a ValueError
        within.expand(start_delta=TimeDelta(-3 * 60, format="sec"))


def test_expand_shrink_negative():
    """Test `expand` method with shrink to negative - should raise `ValueError`."""
    with pytest.raises(ValueError):
        # Test a negative duration shrinkage - this raises a ValueError
        within.expand(
            start_delta=TimeDelta(-5 * 60, format="sec"),
            end_delta=TimeDelta(-5 * 60, format="sec"),
        )


def test_invert(init_times, durations):
    """Test `invert` method."""

    end_times = init_times + durations

    init_intervals = []
    for i, init_time in enumerate(init_times):
        init_intervals.append(TimeInterval(init_times[i], end_times[i]))

    intervals = TimeIntervalList(
        init_intervals,
        start_valid=init_times[0] - TimeDelta(1.0, format="jd"),
        end_valid=end_times[-1] + TimeDelta(1.0, format="jd"),
    )

    inverted_intervals = intervals.invert()

    truth_txt = (
        "[ 2020-04-10T00:00:00.000  2020-04-11T00:00:00.000 ]\n"
        "[ 2020-04-11T00:10:00.000  2020-04-12T00:00:00.000 ]\n"
        "[ 2020-04-12T00:20:00.000  2020-04-13T00:00:00.000 ]\n"
        "[ 2020-04-13T00:30:00.000  2020-04-14T00:30:00.000 ]\n"
    )

    truth_validity = "[ 2020-04-10T00:00:00.000  2020-04-14T00:30:00.000 ]"

    assert truth_txt == str(inverted_intervals)
    assert truth_validity == str(inverted_intervals.valid_interval)



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2021 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Ground and Earth Location classes. These are based on
:class:`astropy.coordinates.EarthLocation` class, but generalised
to non-Earth cases.

"""
import collections

import erfa
import numpy as np
from astropy import units as u
from astropy.coordinates import Angle, Latitude, Longitude

from satmad.core.celestial_bodies import EARTH_ELLIPSOID_GRS80

GeodeticLocation = collections.namedtuple("GeodeticLocation", ["lon", "lat", "height"])


class GroundLocation(u.Quantity):
    """
    Location on a Central Body (e.g. Earth or Moon).

    Initialization is first attempted assuming geocentric (x, y, z) coordinates
    are given; if that fails, another attempt is made assuming geodetic
    coordinates (longitude, latitude, height above a reference ellipsoid).
    When using the geodetic forms, Longitudes are measured increasing to the
    east, so west longitudes are negative. Internally, the coordinates are
    stored as geocentric.

    To ensure a specific type of coordinates is used, use the corresponding
    class methods (`from_geocentric` and `from_geodetic`) or initialize the
    arguments with names (``x``, ``y``, ``z`` for geocentric; ``lon``, ``lat``,
    ``height`` for geodetic).  See the class methods for details.


    Notes
    -----
    This class fits into the coordinates transformation framework in that it
    encodes a position on the `~astropy.coordinates.ITRS` frame.  To get a
    proper `~astropy.coordinates.ITRS` object from this object, use the ``itrs``
    property.
    """

    _location_dtype = np.dtype({"names": ["x", "y", "z"], "formats": [np.float64] * 3})
    _array_dtype = np.dtype((np.float64, (3,)))

    def __new__(cls, *args, **kwargs):

        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], GroundLocation):
            # input is a GroundLocation object, copy it
            return args[0].copy()

        if len(args) == 1 and isinstance(args[0], GeodeticLocation):
            # try parsing as geodetic coordinates (lat, lon, alt)
            self = cls.from_geodetic(
                lon=args[0].lon, lat=args[0].lat, height=args[0].height, **kwargs
            )
            return self

        try:
            # try parsing as geocentric cartesian
            self = cls.from_geocentric(*args, **kwargs)
        except (u.UnitsError, TypeError) as exc_geocentric:
            try:
                # try parsing as geodetic coordinates (lat, lon, alt)
                self = cls.from_geodetic(*args, **kwargs)
            except Exception as exc_geodetic:
                raise TypeError(
                    "Coordinates could not be parsed as either "
                    "geocentric or geodetic, with respective "
                    'exceptions "{}" and "{}"'.format(exc_geocentric, exc_geodetic)
                )
        return self

    @classmethod
    def from_geocentric(cls, x, y, z, unit=None, ellipsoid=EARTH_ELLIPSOID_GRS80):
        """
        Location on Central Body (e.g. Earth or Moon), initialized from
        geocentric cartesian coordinates.

        Parameters
        ----------
        x, y, z : `~astropy.units.Quantity` or array_like
            Cartesian coordinates.  If not quantities, ``unit`` should be given.
        unit : `~astropy.units.UnitBase` object or None
            Physical unit of the coordinate values.  If ``x``, ``y``, and/or
            ``z`` are quantities, they will be converted to this unit.
        ellipsoid : CelestialBodyEllipsoid, optional
            Definition of the reference ellipsoid to use
            (default: 'EARTH_ELLIPSOID_GRS80').

        Raises
        ------
        astropy.units.UnitsError
            If the units on ``x``, ``y``, and ``z`` do not match or an invalid
            unit is given.
        ValueError
            If the shapes of ``x``, ``y``, and ``z`` do not match.
        TypeError
            If ``x`` is not a `~astropy.units.Quantity` and no unit is given.
        """
        if unit is None:
            try:
                unit = x.unit
            except AttributeError:
                raise TypeError(
                    "Geocentric coordinates should be Quantities "
                    "unless an explicit unit is given."
                )
        else:
            unit = u.Unit(unit)

        if unit.physical_type != "length":
            raise u.UnitsError("Geocentric coordinates should be in units of length.")

        try:
            x = u.Quantity(x, unit, copy=False)
            y = u.Quantity(y, unit, copy=False)
            z = u.Quantity(z, unit, copy=False)
        except u.UnitsError:
            raise u.UnitsError("Geocentric coordinate units should all be consistent.")

        x, y, z = np.broadcast_arrays(x, y, z)
        struc = np.empty(x.shape, cls._location_dtype)
        struc["x"], struc["y"], struc["z"] = x, y, z
        self = super().__new__(cls, struc, unit, copy=False)
        self._ellipsoid = ellipsoid
        return self

    @classmethod
    def from_geodetic(cls, lon, lat, height=0.0, ellipsoid=EARTH_ELLIPSOID_GRS80):
        """
        Location on Central Body (e.g. Earth or Moon), initialized from
        geodetic coordinates.

        Parameters
        ----------
        lon : `~astropy.coordinates.Longitude` or float
            Earth East longitude.  Can be anything that initialises an
            `~astropy.coordinates.Angle` object (if float, in degrees).
        lat : `~astropy.coordinates.Latitude` or float
            Earth latitude.  Can be anything that initialises an
            `~astropy.coordinates.Latitude` object (if float, in degrees).
        height : `~astropy.units.Quantity` or float, optional
            Height above reference ellipsoid (if float, in meters; default: 0).
        ellipsoid : CelestialBodyEllipsoid, optional
            Definition of the reference ellipsoid to use
            (default: 'EARTH_ELLIPSOID_GRS80').


        Raises
        ------
        astropy.units.UnitsError
            If the units on ``lon`` and ``lat`` are inconsistent with angular
            ones, or that on ``height`` with a length.
        ValueError
            If ``lon``, ``lat``, and ``height`` do not have the same shape, or
            if ``ellipsoid`` is not recognized as among the ones implemented.

        Notes
        -----
        For the conversion to geocentric coordinates, the ERFA routine
        ``gd2gce`` is used.  See https://github.com/liberfa/erfa
        """
        # We use Angle here since there is no need to wrap the longitude -
        # gd2gce will just take cos/sin anyway.  And wrapping might fail
        # on readonly input.
        lon = Angle(lon, u.degree, copy=False)
        lat = Latitude(lat, u.degree, copy=False)
        # don't convert to m by default, so we can use the height unit below.
        if not isinstance(height, u.Quantity):
            height = u.Quantity(height, u.m, copy=False)
        # get geocentric coordinates. Have to give one-dimensional array.
        xyz = erfa.gd2gce(
            ellipsoid.re.to_value(u.m),
            1 / ellipsoid.inv_f.to_value(),
            lon.to_value(u.radian),
            lat.to_value(u.radian),
            height.to_value(u.m),
        )
        self = xyz.ravel().view(cls._location_dtype, cls).reshape(xyz.shape[:-1])
        self._unit = u.m
        self._ellipsoid = ellipsoid
        return self

    @property
    def ellipsoid(self):
        """The default ellipsoid used to convert to geodetic coordinates."""
        return self._ellipsoid

    @property
    def geodetic(self):
        """Convert to geodetic coordinates for the default ellipsoid."""
        return self.to_geodetic()

    def to_geodetic(self, ellipsoid=EARTH_ELLIPSOID_GRS80):
        """Convert to geodetic coordinates.

        Parameters
        ----------
        ellipsoid : CelestialBodyEllipsoid, optional
            Definition of the reference ellipsoid to use
            (default: 'EARTH_ELLIPSOID_GRS80').

        Returns
        -------
        (lon, lat, height) : tuple
            The tuple contains instances of `~astropy.coordinates.Longitude`,
            `~astropy.coordinates.Latitude`, and `~astropy.units.Quantity`


        Notes
        -----
        For the conversion to geodetic coordinates, the ERFA routine
        ``gc2gde`` is used.  See https://github.com/liberfa/erfa
        """

        self_array = self.to(u.meter).view(self._array_dtype, np.ndarray)
        lon, lat, height = erfa.gc2gde(
            ellipsoid.re.to_value(u.m), 1 / ellipsoid.inv_f.to_value(), self_array
        )
        return GeodeticLocation(
            Longitude(
                lon * u.radian, u.degree, wrap_angle=180.0 * u.degree, copy=False
            ),
            Latitude(lat * u.radian, u.degree, copy=False),
            u.Quantity(height * u.meter, self.unit, copy=False),
        )

    @property
    def lon(self):
        """Longitude of the location, for the default ellipsoid."""
        return self.geodetic[0]

    @property
    def lat(self):
        """Longitude of the location, for the default ellipsoid."""
        return self.geodetic[1]

    @property
    def height(self):
        """Height of the location, for the default ellipsoid."""
        return self.geodetic[2]

    # mostly for symmetry with geodetic and to_geodetic.
    @property
    def geocentric(self):
        """Convert to a tuple with X, Y, and Z as quantities"""
        return self.to_geocentric()

    def to_geocentric(self):
        """Convert to a tuple with X, Y, and Z as quantities"""
        return self.x, self.y, self.z

    def to_body_fixed_coords(self, celestial_body, obstime=None):
        """
        Generates a Central Body Fixed object (e.g. `~astropy.coordinates.ITRS` object)
        with the location of this object at the requested ``obstime``.

        The `celestial_body` object should have its `body_fixed_coord` defined, as the
        resulting object will be of this type. For example, if the celestial body is
        Earth, the resulting object is of the type ITRS.

        Parameters
        ----------
        celestial_body : CelestialBody
            The Celestial Body where the object is defined.
        obstime : `~astropy.time.Time` or None
            The ``obstime`` to apply to the new object, or
            if None, the default ``obstime`` will be used.

        Returns
        -------
        body_fixed
            The new object in the body fixed frame

        Raises
        ------

        """
        # Broadcast for a single position at multiple times, but don't attempt
        # to be more general here.
        if obstime and self.size == 1 and obstime.shape:
            self = np.broadcast_to(self, obstime.shape, subok=True)

        try:
            return celestial_body.body_fixed_coord_frame(
                x=self.x, y=self.y, z=self.z, obstime=obstime
            )
        except TypeError:
            raise TypeError(
                "Failed converting Geodetic coordinates to body fixed coordinates. "
                "Celestial Body does not have a body fixed coordinate system defined, "
                "e.g. ITRS for Earth."
            )

    @property
    def x(self):
        """The X component of the geocentric coordinates."""
        return self["x"]

    @property
    def y(self):
        """The Y component of the geocentric coordinates."""
        return self["y"]

    @property
    def z(self):
        """The Z component of the geocentric coordinates."""
        return self["z"]

    def __getitem__(self, item):
        result = super().__getitem__(item)
        if result.dtype is self.dtype:
            return result.view(self.__class__)
        else:
            return result.view(u.Quantity)

    def __array_finalize__(self, obj):
        super().__array_finalize__(obj)
        if hasattr(obj, "_ellipsoid"):
            self._ellipsoid = obj.ellipsoid

    def __len__(self):
        if self.shape == ():
            raise IndexError("0-d GroundLocation arrays cannot be indexed")
        else:
            return super().__len__()

    def _to_value(self, unit, equivalencies=None):
        """Helper method for to and to_value."""
        # Conversion to another unit in both ``to`` and ``to_value`` goes
        # via this routine. To make the regular quantity routines work, we
        # temporarily turn the structured array into a regular one.
        if equivalencies is None:
            equivalencies = []
        array_view = self.view(self._array_dtype, np.ndarray)
        if not equivalencies:
            equivalencies = self._equivalencies
        new_array = self.unit.to(unit, array_view, equivalencies=equivalencies)
        return new_array.view(self.dtype).reshape(self.shape)



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2021 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
TLE Storage class tests

"""
from pathlib import Path

import pytest
from astropy import units as u
from astropy.time import Time

from satmad.propagation.tle_storage import TleFilterParams, TleStorage, TleTimeSeries

extra_path = Path("satmad", "propagation", "tests")

mixed_tle_file_path_1 = Path("data", "tle_mixed_1.txt")
time_series_tle_file_path_1 = Path("data", "tle_rasat_desc.txt")
time_series_tle_file_path_2 = Path("data", "tle_rasat.txt")


def test_parse_storage_file():
    """Test parsing of the storage file with mixed TLE input."""
    file_path = process_paths(mixed_tle_file_path_1)

    tle_storage = TleStorage.from_path(file_path)

    truth_no_of_elements = 19
    truth_pos = 3
    truth_name = "APRIZESAT 1"
    truth_line1 = (
        "1 28372U 04025G   21088.15052337  .00000024  00000-0  18112-4 0  9991"
    )
    truth_line2 = (
        "2 28372  98.4324  42.3607 0046090 190.1598 169.8668 14.49665428885585"
    )

    # check number of elements
    assert len(tle_storage.tle_list) == truth_no_of_elements

    # check specific element at `truth_pos`
    assert str(tle_storage.tle_list[truth_pos]).split("\n")[0:3] == [
        truth_name,
        truth_line1,
        truth_line2,
    ]


def test_tle_param_filter_error():
    with pytest.raises(ValueError):
        """Tests filtering for a TleFilterParams.TLE error."""
        file_path = process_paths(mixed_tle_file_path_1)

        tle_storage = TleStorage.from_path(file_path)

        tle_storage.filter_by_value(TleFilterParams.TLE, 46495)


def test_filter_value():
    """Tests filtering for a value equivalence."""
    file_path = process_paths(mixed_tle_file_path_1)

    tle_storage = TleStorage.from_path(file_path)

    filtered_list_sat_nr = tle_storage.filter_by_value(
        TleFilterParams.SAT_NUMBER, 46495
    )
    filtered_list_int_deg = tle_storage.filter_by_value(
        TleFilterParams.INTL_DESIGNATOR, "18014H"
    )
    empty_filtered_list = tle_storage.filter_by_value(TleFilterParams.SAT_NUMBER, 56495)

    assert isinstance(filtered_list_sat_nr, TleStorage)
    assert tle_storage.tle_list[0].name == "APRIZESAT 2"
    assert len(filtered_list_sat_nr.tle_list) == 2
    assert len(filtered_list_int_deg.tle_list) == 1
    assert len(empty_filtered_list.tle_list) == 0


def test_filter_func():
    """Tests filtering for a value range with a function."""
    file_path = process_paths(mixed_tle_file_path_1)

    tle_storage = TleStorage.from_path(file_path)

    def sma_filter_1(a):
        """Semimajor axis filter min."""
        return True if a > 7000 * u.km else False

    def sma_filter_2(a):
        """Semimajor axis filter max."""
        return True if 7000 * u.km > a else False

    def sma_filter_3(a):
        """Semimajor axis filter min/max."""
        return True if 7100 * u.km > a > 7000 * u.km else False

    def tle_filter_1(tle):
        """Semimajor axis filter min/max."""
        return True if 7100 * u.km > tle.sm_axis > 7000 * u.km else False

    filtered_list_sma_1 = tle_storage.filter_by_func(
        TleFilterParams.SM_AXIS, sma_filter_1
    )
    filtered_list_sma_2 = tle_storage.filter_by_func(
        TleFilterParams.SM_AXIS, sma_filter_2
    )
    filtered_list_sma_3 = tle_storage.filter_by_func(
        TleFilterParams.SM_AXIS, sma_filter_3
    )
    filtered_list_sma_4 = tle_storage.filter_by_func(TleFilterParams.TLE, tle_filter_1)

    assert len(filtered_list_sma_1.tle_list) == 11
    assert len(filtered_list_sma_2.tle_list) == 8
    assert len(filtered_list_sma_3.tle_list) == 8
    assert len(filtered_list_sma_4.tle_list) == 8


def test_filter_range():
    """Tests filtering for a value range with min/max parameters."""
    file_path = process_paths(mixed_tle_file_path_1)

    tle_storage = TleStorage.from_path(file_path)

    filtered_list_sma_1 = tle_storage.filter_by_range(
        TleFilterParams.SM_AXIS, min_value=7000 * u.km
    )

    filtered_list_sma_2 = tle_storage.filter_by_range(
        TleFilterParams.SM_AXIS, max_value=7000 * u.km
    )

    filtered_list_sma_3 = tle_storage.filter_by_range(
        TleFilterParams.SM_AXIS, min_value=7000 * u.km, max_value=7100 * u.km
    )

    filtered_list_sma_4 = tle_storage.filter_by_range(TleFilterParams.SM_AXIS)

    assert len(filtered_list_sma_1.tle_list) == 11
    assert len(filtered_list_sma_2.tle_list) == 8
    assert len(filtered_list_sma_3.tle_list) == 8
    assert len(filtered_list_sma_4.tle_list) == 0


def test_tle_timeseries_ordered():
    """Test parsing of the TLE Timeseries with ordered time input."""
    file_path = process_paths(time_series_tle_file_path_2)

    tle_storage = TleStorage.from_path(file_path).to_tle_timeseries(37791)

    assert tle_storage.tle_list[0].epoch.isot == "2021-03-15T02:02:02.753"
    assert tle_storage.tle_list[-1].epoch.isot == "2021-03-31T21:19:42.808"


def test_tle_timeseries_unordered():
    """Test parsing of the TLE Timeseries with inverted time input."""
    file_path = process_paths(time_series_tle_file_path_1)

    tle_storage = TleStorage.from_path(file_path).to_tle_timeseries(37791)

    assert tle_storage.tle_list[0].epoch.isot == "2021-03-30T04:20:36.404"
    assert tle_storage.tle_list[-1].epoch.isot == "2021-04-01T20:16:48.785"


def test_tle_timeseries_time_filter():
    """Test parsing of the TLE Timeseries with time filter."""
    file_path = process_paths(time_series_tle_file_path_1)

    tle_storage = TleStorage.from_path(file_path).to_tle_timeseries(37791)

    threshold_time = Time("2021-04-01T00:00:00.000")
    tle_storage_filtered = tle_storage.filter_by_range(
        TleFilterParams.EPOCH, min_value=threshold_time
    )

    assert isinstance(tle_storage_filtered, TleTimeSeries)
    assert tle_storage_filtered.tle_list[0].epoch.isot == "2021-04-01T02:14:48.404"
    assert tle_storage_filtered.tle_list[-1].epoch.isot == "2021-04-01T20:16:48.785"


def process_paths(path):
    """
    Processes the path depending on the run environment.
    """
    working_dir = Path.cwd()

    file_path = working_dir.joinpath(path)
    if not working_dir.joinpath(file_path).exists():
        file_path = working_dir.joinpath(extra_path).joinpath(path)

    return file_path



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['satmad',
 'satmad.coordinates',
 'satmad.coordinates.tests',
 'satmad.core',
 'satmad.core.tests',
 'satmad.propagation',
 'satmad.propagation.tests',
 'satmad.tests',
 'satmad.utils',
 'satmad.utils.tests']

package_data = \
{'': ['*'],
 'satmad': ['data/*', 'temp/*'],
 'satmad.coordinates': ['data/*'],
 'satmad.coordinates.tests': ['.pytest_cache/*',
                              '.pytest_cache/v/*',
                              '.pytest_cache/v/cache/*'],
 'satmad.propagation.tests': ['data/*'],
 'satmad.tests': ['data/*'],
 'satmad.utils.tests': ['.pytest_cache/*',
                        '.pytest_cache/v/*',
                        '.pytest_cache/v/cache/*']}

install_requires = \
['astropy', 'matplotlib', 'numpy', 'portion', 'scipy', 'sgp4']

extras_require = \
{'doc': ['sphinx'], 'test': ['pytest', 'pytest-cov', 'pytest-xdist']}

setup(name='satmad',
      version='0.1',
      description='satmad',
      author='Egemen Imre',
      author_email=None,
      url='https://github.com/egemenimre/satmad/',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) ${YEAR} Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Test coordinate transformations for the TEME, J2000 and TIRS coordinate frames.

"""
from astropy import units as u
from astropy.coordinates import (
    CIRS,
    GCRS,
    HCRS,
    ICRS,
    ITRS,
    TEME,
    CartesianDifferential,
    CartesianRepresentation,
    SkyCoord,
)
from astropy.time import Time
from pytest import approx

from satmad.coordinates.frames import J2000, TIRS, CelestialBodyCRS
from satmad.core.celestial_bodies import EARTH, SUN

time: Time = Time("2004-04-06T07:51:28.386009", scale="utc")

# Vallado IAU 2000 - Table 3-6
v_gcrs_true = CartesianDifferential(
    [-4.7432201610, 0.7905364950, 5.5337557240], unit=u.km / u.s
)
r_gcrs_true = CartesianRepresentation(
    [5102.50895290, 6123.01139910, 6378.13693380], unit=u.km
)
rv_gcrs_true = GCRS(
    r_gcrs_true.with_differentials(v_gcrs_true),
    obstime=time,
    representation_type="cartesian",
    differential_type="cartesian",
)

# Vallado IAU 2000 - Table 3-6 J2000
v_j2000_true = CartesianDifferential(
    [-4.7432196000, 0.7905366000, 5.5337561900], unit=u.km / u.s
)
r_j2000_true = CartesianRepresentation(
    [5102.50960000, 6123.01152000, 6378.13630000], unit=u.km
)
rv_j2000_true = J2000(
    r_j2000_true.with_differentials(v_j2000_true),
    obstime=time,
    representation_type="cartesian",
    differential_type="cartesian",
)
# Vallado IAU 2000 - Table 3-6 CIRS
v_cirs_true = CartesianDifferential(
    [-4.7453803300, 0.7903414530, 5.5319312880], unit=u.km / u.s
)
r_cirs_true = CartesianRepresentation(
    [5100.01840470, 6122.78636480, 6380.34453270], unit=u.km
)
rv_cirs_true = CIRS(
    r_cirs_true.with_differentials(v_cirs_true),
    obstime=time,
    representation_type="cartesian",
    differential_type="cartesian",
)

# Vallado IAU 2000 - Table 3-6 ITRS
v_itrs_true = CartesianDifferential(
    [-3.2256365200, -2.8724514500, 5.5319244460], unit=u.km / u.s
)
r_itrs_true = CartesianRepresentation(
    [-1033.4793830, 7901.29527540, 6380.35659580], unit=u.km
)
rv_itrs_true = ITRS(
    r_itrs_true.with_differentials(v_itrs_true),
    obstime=time,
    representation_type="cartesian",
    differential_type="cartesian",
)

# Vallado IAU 2000 - Table 3-6 TIRS
v_tirs_true = CartesianDifferential(
    [-3.2256327470, -2.8724425110, 5.5319312880], unit=u.km / u.s
)
r_tirs_true = CartesianRepresentation(
    [-1033.47503120, 7901.30558560, 6380.34453270], unit=u.km
)
rv_tirs_true = TIRS(
    r_tirs_true.with_differentials(v_tirs_true),
    obstime=time,
    representation_type="cartesian",
    differential_type="cartesian",
)

# Vallado IAU 2000 - Table 3-6 TEME
v_teme_true = CartesianDifferential(
    [-4.7461314870, 0.7858180410, 5.5319312880], unit=u.km / u.s
)
r_teme_true = CartesianRepresentation(
    [5094.18016210, 6127.64465950, 6380.34453270], unit=u.km
)
rv_teme_true = TEME(
    r_teme_true.with_differentials(v_teme_true),
    obstime=time,
    representation_type="cartesian",
    differential_type="cartesian",
)


def pos_err(rv_test, rv_true):
    """
    Computes positional error between two vectors defined in a frame.

    Parameters
    ----------
    rv_test : State with position vector under test
    rv_true : State with true position vector

    Returns
    -------
    Quantity
        3D position difference

    """
    r_diff = (
        rv_test.cartesian.without_differentials()
        - rv_true.cartesian.without_differentials()
    )
    return r_diff.norm().to(u.mm)


def vel_err(rv_test, rv_true):
    """
    Computes velocity error between two vectors defined in a frame.

    Parameters
    ----------
    rv_test : State with velocity vector under test
    rv_true : State with true velocity vector

    Returns
    -------
    Quantity
        3D velocity difference

    """
    v_diff = rv_test.velocity - rv_true.velocity
    return v_diff.norm().to(u.mm / u.s)


# ********** Functional testing **********


def test_tirs_to_teme_no_vel():
    """Check whether coord transform without velocity is possible."""
    rv_tirs_no_vel = TIRS(r_tirs_true, obstime=time, representation_type="cartesian")
    rv_tirs_no_vel.transform_to(TEME(obstime=time))


def test_teme_to_tirs_no_vel():
    """Check whether coord transform without velocity is possible."""
    rv_teme_no_vel = TEME(r_teme_true, obstime=time, representation_type="cartesian")
    rv_teme_no_vel.transform_to(TIRS(obstime=time))


def test_itrs_roundtrip():
    """Check whether transforming to a coord and then
    transforming back yields the same output."""
    # test_frame = "ITRS"
    allowable_pos_diff = 1.5e-6 * u.mm
    allowable_vel_diff = 2.8e-6 * u.mm / u.s

    rv_teme = rv_itrs_true.transform_to(TEME(obstime=time))
    rv_itrs_from_teme = rv_teme.transform_to(ITRS(obstime=time))

    r_diff = pos_err(rv_itrs_from_teme, rv_itrs_true)
    v_diff = vel_err(rv_itrs_from_teme, rv_itrs_true)

    # print(f"r {rv_itrs_true.name} diff      :  {r_diff}")
    # print(f"v {rv_itrs_true.name} diff      :  {v_diff}")

    assert approx(r_diff.value, abs=allowable_pos_diff.value) == 0.0
    assert approx(v_diff.value, abs=allowable_vel_diff.value) == 0.0


def test_j2000_roundtrip():
    """Check whether transforming to a coord and then
    transforming back yields the same output."""
    # test_frame = "J2000"
    allowable_pos_diff = 1.5e-6 * u.mm
    allowable_vel_diff = 1.0e-9 * u.mm / u.s

    rv_gcrs = rv_j2000_true.transform_to(GCRS(obstime=time))
    rv_j2000_from_gcrs = rv_gcrs.transform_to(J2000(obstime=time))

    r_diff = pos_err(rv_j2000_from_gcrs, rv_j2000_true)
    v_diff = vel_err(rv_j2000_from_gcrs, rv_j2000_true)

    # print(f"r {test_frame} diff      :  {r_diff}")
    # print(f"v {test_frame} diff      :  {v_diff}")

    assert approx(r_diff.value, abs=allowable_pos_diff.value) == 0.0
    assert approx(v_diff.value, abs=allowable_vel_diff.value) == 0.0


# ********** Performance testing **********


def test_j2000_to_gcrs():
    """Check the coordinate transform accuracy."""
    # test_frame = "GCRS"
    allowable_pos_diff = 800 * u.mm
    allowable_vel_diff = 0.36 * u.mm / u.s

    rv_gcrs = rv_j2000_true.transform_to(GCRS(obstime=time))

    r_diff = pos_err(rv_gcrs, rv_gcrs_true)
    v_diff = vel_err(rv_gcrs, rv_gcrs_true)

    # print(f"r {test_frame} diff      :  {r_diff}")
    # print(f"v {test_frame} diff      :  {v_diff}")

    assert approx(r_diff.value, abs=allowable_pos_diff.value) == 0.0
    assert approx(v_diff.value, abs=allowable_vel_diff.value) == 0.0


def test_teme_to_tirs():
    """Check the coordinate transform accuracy."""
    # test_frame = "TIRS"
    allowable_pos_diff = 300 * u.mm
    allowable_vel_diff = 0.20 * u.mm / u.s

    rv_tirs = rv_teme_true.transform_to(TIRS(obstime=time))

    r_diff = pos_err(rv_tirs, rv_tirs_true)
    v_diff = vel_err(rv_tirs, rv_tirs_true)

    # print(f"r {rv_tirs.name} diff      :  {r_diff}")
    # print(f"v {rv_tirs.name} diff      :  {v_diff}")

    assert approx(r_diff.value, abs=allowable_pos_diff.value) == 0.0
    assert approx(v_diff.value, abs=allowable_vel_diff.value) == 0.0


def test_itrs_to_teme():
    """Check the coordinate transform accuracy."""
    # test_frame = "TEME"
    allowable_pos_diff = 300 * u.mm
    allowable_vel_diff = 0.23 * u.mm / u.s

    rv_teme = rv_itrs_true.transform_to(TEME(obstime=time))

    r_diff = pos_err(rv_teme, rv_teme_true)
    v_diff = vel_err(rv_teme, rv_teme_true)

    # print(f"r {rv_teme.name} diff      :  {r_diff}")
    # print(f"v {rv_teme.name} diff      :  {v_diff}")

    assert approx(r_diff.value, abs=allowable_pos_diff.value) == 0.0
    assert approx(v_diff.value, abs=allowable_vel_diff.value) == 0.0


def test_itrs_to_tirs():
    """Check the coordinate transform accuracy."""
    # test_frame = "TIRS"
    allowable_pos_diff = 60 * u.mm
    allowable_vel_diff = 0.05 * u.mm / u.s

    rv_tirs = rv_itrs_true.transform_to(TIRS(obstime=time))

    r_diff = pos_err(rv_tirs, rv_tirs_true)
    v_diff = vel_err(rv_tirs, rv_tirs_true)

    # print(f"r {test_frame} diff      :  {r_diff}")
    # print(f"v {test_frame} diff      :  {v_diff}")

    assert approx(r_diff.value, abs=allowable_pos_diff.value) == 0.0
    assert approx(v_diff.value, abs=allowable_vel_diff.value) == 0.0


def test_tirs_to_itrs():
    """Check the coordinate transform accuracy."""
    # test_frame = "ITRS"
    allowable_pos_diff = 60 * u.mm
    allowable_vel_diff = 0.05 * u.mm / u.s

    rv_itrs = rv_tirs_true.transform_to(ITRS(obstime=time))

    r_diff = pos_err(rv_itrs, rv_itrs_true)
    v_diff = vel_err(rv_itrs, rv_itrs_true)

    # print(f"r {test_frame} diff      :  {r_diff}")
    # print(f"v {test_frame} diff      :  {v_diff}")

    assert approx(r_diff.value, abs=allowable_pos_diff.value) == 0.0
    assert approx(v_diff.value, abs=allowable_vel_diff.value) == 0.0


class _EarthCRS(CelestialBodyCRS):
    body = EARTH


class _SunCRS(CelestialBodyCRS):
    body = SUN


def test_cb_crs_to_icrs_earth():
    """Check the basic quasi-round-trip accuracy for the generic Central Body Celestial
    Reference System, using Earth (equal to GCRS). Converts GCRS to ICRS, then ICRS to
    Earth CRS. However there is a difference, possibly due to the difference between
    the two transformations (`FunctionTransformWithFiniteDifference` vs
    `AffineTransform`)"""
    allowable_pos_diff = 800 * u.m
    allowable_vel_diff = 0.510 * u.m / u.s

    rv_icrs = SkyCoord(rv_gcrs_true).transform_to(ICRS)

    rv_gcrs = rv_icrs.transform_to(_EarthCRS(obstime=rv_gcrs_true.obstime))

    r_diff = pos_err(rv_gcrs, rv_gcrs_true)
    v_diff = vel_err(rv_gcrs, rv_gcrs_true)

    # print(f"r {rv_gcrs.name} diff      :  {r_diff}")
    # print(f"v {rv_gcrs.name} diff      :  {v_diff}")

    assert approx(r_diff.to_value(u.mm), abs=allowable_pos_diff.to_value(u.mm)) == 0.0
    assert (
        approx(v_diff.to_value(u.mm / u.s), abs=allowable_vel_diff.to_value(u.mm / u.s))
        == 0.0
    )


def test_cb_crs_to_icrs_sun():
    """Check the basic quasi-round-trip accuracy for the generic Central Body Celestial
    Reference System, using Sun (equal to HCRS). Converts Sun CRS to ICRS, then ICRS to
    HCRS."""
    allowable_pos_diff = 5e-4 * u.mm
    allowable_vel_diff = 0.0001 * u.mm / u.s

    rv_suncrs_true = SkyCoord(
        r_gcrs_true.with_differentials(v_gcrs_true),
        obstime=time,
        frame=_SunCRS,
        representation_type="cartesian",
        differential_type="cartesian",
    )

    rv_icrs = rv_suncrs_true.transform_to(ICRS)

    rv_hcrs = rv_icrs.transform_to(HCRS(obstime=rv_suncrs_true.obstime))

    r_diff = pos_err(rv_hcrs, rv_suncrs_true)
    v_diff = vel_err(rv_hcrs, rv_suncrs_true)

    # print(f"r {rv_hcrs.name} diff      :  {r_diff}")
    # print(f"v {rv_hcrs.name} diff      :  {v_diff}")

    assert approx(r_diff.to_value(u.mm), abs=allowable_pos_diff.to_value(u.mm)) == 0.0
    assert (
        approx(v_diff.to_value(u.mm / u.s), abs=allowable_vel_diff.to_value(u.mm / u.s))
        == 0.0
    )



# SatMAD: Satellite Mission Analysis and Design for Python
#
# Copyright (C) 2020 Egemen Imre
#
# Licensed under GNU GPL v3.0. See LICENSE.rst for more info.
"""
Discrete time events and intervals finding methods.

"""
from astropy import units as u
from astropy.timeseries import TimeSeries
from scipy import interpolate

from satmad.utils.timeinterval import TimeInterval, TimeIntervalList


class DiscreteTimeEvents:
    """
    Discrete time events and intervals finding and storage class.

    An *time event* is described as a function value crossing a certain threshold in
    increasing (negative to positive) and decreasing (positive to negative) directions,
    defining a time interval.

    The class initialises interpolators to find where the `value_list` crosses the
    `crossing_value`, which is essentially a root finding problem.

    Parameters
    ----------
    time_list : Time
        List of time instances
    value_list : ndarray or list
        List of values corresponding to time instances (should be of the same size as
        `time_list`)
    crossing_value : float
        Crossing value for the `value_list`
    neg_to_pos_is_start : bool
        If `True` value turning from negative to positive marks a *start* event,
        otherwise marks an *end* event

    Raises
    ------
    ValueError
        `time_list` and `value_list` are of different sizes
    """

    def __init__(
        self, time_list, value_list, crossing_value=0, neg_to_pos_is_start=True
    ):

        # Check whether time_list and value_list size are equal
        if len(time_list) != len(value_list):
            raise ValueError(
                f"Time list and value list are of different sizes. "
                f"({len(time_list)} vs. ({len(value_list)}))"
            )

        self.search_interval = TimeInterval(time_list[0], time_list[-1])

        # Convert time to "days since epoch"
        t_float_list = (time_list - time_list[0]).jd

        # Init interpolators (for splines, root finding possible for 3rd degree
        # (cubics) only)
        self._interpolator = interpolate.CubicSpline(
            t_float_list, value_list - crossing_value
        )

        self._deriv_interpolator = self._interpolator.derivative()

        # Find start / end intervals
        self.start_end_intervals = self._find_intervals(time_list, neg_to_pos_is_start)

        # Find max / min events
        self.max_min_table = self._find_extrema_events(time_list[0])

        # Return the table to absolute values
        self.max_min_table["value"] = self.max_min_table["value"] + crossing_value

    def _find_intervals(self, time_list, neg_to_pos_is_start):
        """
        Finds the start / end intervals within the time list.

        Parameters
        ----------
        time_list : Time
            List of time instances
        neg_to_pos_is_start : bool
            If `True` value turning from negative to positive marks a *start* event,
            otherwise marks an *end* event

        Returns
        -------
        intervals : TimeIntervalList
            Time intervals marked with start and end events
        """

        init_time = time_list[0]
        end_time = time_list[-1]

        # *** Find start / end events (do not extrapolate to find roots) ***
        # No roots mean no events during this time
        start_end_events = self._interpolator.roots(extrapolate=False)

        intervals = []
        if start_end_events.size != 0:
            # there should be some roots within the search range to check for intervals

            # classify start / end events and fill the timetable
            events = self._classify_start_end_events(
                init_time, start_end_events, neg_to_pos_is_start
            )

            # Put events into intervals
            event_start = init_time

            for event in events:
                # ignore events outside the search interval
                if self.search_interval.is_in_interval(event["time"]):
                    if event["type"] == "start":
                        event_start = event["time"]
                    else:
                        intervals.append(TimeInterval(event_start, event["time"]))

            # check for the last event - force end time if it is out of search bounds
            if events[-1]["type"] == "end":
                if self.search_interval.is_in_interval(events[-1]["time"]):
                    intervals.append(TimeInterval(event_start, events[-1]["time"]))
                else:
                    intervals.append(
                        TimeInterval(event_start, self.search_interval.end)
                    )
            else:
                intervals.append(TimeInterval(event_start, self.search_interval.end))

        # If interval list is empty by this stage, either the complete duration is
        # is a valid interval with no event (e.g. a continuously visible satellite)
        # or the complete duration is an invalid interval with no event (e.g. a
        # satellite continuously outside visibility)
        if not intervals:
            mid_time = 0.5 * (self.search_interval.end - self.search_interval.start)
            value = self._interpolator(mid_time.to_value(u.day))

            if neg_to_pos_is_start:
                if value > 0:
                    # event already started
                    intervals.append(
                        TimeInterval(
                            self.search_interval.start, self.search_interval.end
                        )
                    )
            else:
                if value < 0:
                    # event already started
                    intervals.append(
                        TimeInterval(
                            self.search_interval.start, self.search_interval.end
                        )
                    )

        # generate the list of time intervals
        return TimeIntervalList(intervals, start_valid=init_time, end_valid=end_time)

    def _classify_start_end_events(
        self, init_time, start_end_events, neg_to_pos_is_start
    ):
        """
        Classify the list of events into a timetable of start and end events.

        Parameters
        ----------
        init_time : Time
            Initial time
        start_end_events : ndarray
            Array of start and end event times in days, starting from `init_time`
        neg_to_pos_is_start : bool
            If `True` value turning from negative to positive marks a *start* event,
            otherwise marks an *end* event

        Returns
        -------
        events_table : TimeSeries
            Timetable of start and end events
        """

        # loop through each root and classify them as start / end events
        events_list = []
        for event_time in start_end_events:

            deriv = self._deriv_interpolator(event_time)

            if neg_to_pos_is_start:
                if deriv > 0:
                    events_list.append("start")
                else:
                    events_list.append("end")
            else:
                if deriv < 0:
                    events_list.append("start")
                else:
                    events_list.append("end")

        # init events table
        events_table = TimeSeries(
            time=init_time + start_end_events * u.day, data={"type": events_list}
        )

        return events_table

    def _find_extrema_events(self, init_time):
        """
        Finds the extrema (max/min) events, contained within the time intervals.

        Parameters
        ----------
        init_time : Time
            Initial time

        Returns
        -------
        max_min_table : TimeSeries
            Timetable of max and min events

        """

        # *** Find max / min events ***
        min_max_events = self._deriv_interpolator.roots()

        # loop through each root and classify them as max / min events
        events_list = []
        value_list = []
        for event_time in min_max_events:
            value = self._interpolator(event_time)
            value_list.append(value)
            if value > 0:
                events_list.append("max")
            else:
                events_list.append("min")

        # init events table
        max_min_table = TimeSeries(
            time=init_time + min_max_events * u.day,
            data={"type": events_list, "value": value_list},
        )

        # Filter the events to those within the intervals
        i = 0
        remove_indexes = []
        for event_row in max_min_table:
            event_within_interval = self.start_end_intervals.is_in_interval(
                event_row["time"]
            )
            if not event_within_interval:
                remove_indexes.append(i)
            i = i + 1

        # Remove the ones outside the intervals
        max_min_table.remove_rows(remove_indexes)

        return max_min_table
